%% Dbut du module echo_server
-module(echo_server).

-export([start/0, start/1, stop/1]).
-export([server/1, client_manager/1]).

-define(default_port, 9999).

%% Dmarrage du processus serveur sur le port par dfaut
start() ->
    start(?default_port).

%% Dmarrage du processus serveur sur le port pass en paramtre
start(Port) ->
    %% Notre programme coute sur le port Port 
    %% Nous rcuprons la socket sur lequel notre programme coute
    %% pour nous mettre ensuite en attente de connexion
    %% Reuseaddr permet de relancer rapidement un serveur aprs son arret et de pouvoir rutiliser immdiatement l'adresse qu'il a libr
    SocketDEcoute = case gen_tcp:listen(Port, [{packet,0}, {active, false}, {reuseaddr, true}]) of
                        %% L'accs au port demand est possible
                        {ok, LSock} -> io:format("Serveur dmarr sur le port ~p~n", [Port]),
                                       LSock;
                        %% Ou bien une erreur s'est produite:
                        {error, Raison} -> io:format("Erreur dans le dmarrage du serveur sur le port ~p: [~p]~n", 
                                                     [Port, Raison]),
                                           exit(Raison)
                    end,
    spawn(?MODULE, server, [SocketDEcoute]).

stop(Pid) ->
    Pid ! {stop},
    ok.

server(LSock) ->
    %% Rcupre les messages de fin de processus, pour tre capable d'afficher sur l'cran lorsqu'un processus se dconnecte: On positionne l'indicateur du processus pour qu'il capture les messages de sortie (EXIT).
    process_flag(trap_exit, true),
    loop(LSock).

%% Le processus serveur
loop(LSock) ->
    %% Le processus serveur accepte les connexions. Pour chaque
    %% nouvelle connexion, il rcupre la socket de connexion avec le
    %% client (Diffrente de la socket qui sert  coute pour les
    %% nouvelles connexions entrantes)
    %% Dans un deuxime temps, il cre un processus pour grer le
    %% nouveau client et lui transmet le contrle de la socket de
    %% connexion avec le client.
    %% Le timeout permet dans l'accept permet de traiter proprement
    %% l'arrt du serveur: L'accept n'est plus bloquant: En itrant 
    %% nouveau sur la boucle le serveur peut excuter d'autres
    %% traitements: Rception du message d'arrt du serveur, mais
    %% galement rception des informations concernant la dconnexion
    %% des clients.
 
    %% TODO: A mettre dans une autre fonction pour plus de clart ?
    receive
        {stop} ->
            io:format("Serveur arrt~n", []),
            exit(normal);
        {'EXIT', Pid, Raison} ->
            io:format("Un client vient de se dconnecter~n", [])
    after 100 ->
            timeout
    end,

    case gen_tcp:accept(LSock, 100) of
        {ok, Sock} ->
            io:format("Connexion d'un nouveau client~n", []),
    
    %% Lancement du processus de gestion de la connexion cliente et
    %% rcupration de son indentifiant de processus

    %% Il faut mettre spawn_link pour pouvoir arrter les
    %% connexions clientes en mme tant que le process serveur
    %% principal.
    %% En plus d'utiliser spawn_link il faut traiter les informations
    %% sur les processus client comme un message, plutt pour viter
    %% le ct bidirectionnel du spawn_link: L'arrt du serveur
    %% entraine l'arrt de tous les processus clients, alors que
    %% l'arrt d'un client n'entraine pas l'arrt du serveur. Il est
    %% trait par le serveur.
            PClientID = spawn_link(?MODULE, client_manager, [Sock]),

    %% Cette commande technique est indispensable, pour transfrer le
    %% contrle de la socket de connexion au nouveau server.
            gen_tcp:controlling_process(Sock, PClientID),
            server(LSock);
        %% Laisse au server l'opportunit de recevoir des messages
        {error, timeout} ->
            server(LSock)
    end.

client_manager(Socket) ->
    client_manager(Socket, []).
client_manager(Socket, Tampon) ->
    case gen_tcp:recv(Socket, 0) of
        {ok, Donnees} ->
            Reste_A_Traiter = action(Socket, lists:flatten([Tampon,Donnees])),
            client_manager(Socket, Reste_A_Traiter);
        {error, closed} ->
            gen_tcp:close(Socket)
    end.

%% Mettre une explication de pourquoi stocker les donnes
%% intermdiaires: Explication sur le fonctionnement de TCP/IP.
%% Montrer ventuellement comment programmer le serveur plus simplement.

%% TODO: La commande: ok = gen_tcp:close(Sock), Ne s'applique pas sur
%% la socket d'coute, mais sur la socket de connexion avec le client.

%% 1. Process d'attente de connexion.

%% 2. Processus de gestion d'une connexion
%% => Un processus de ce type va tre cr pour chaque nouvelle
%% connexion: Il y autant de ces processus tournant simultanment sur
%% le systme qu'il y a de clients simultanment connect.

%% Cette fonction rcupre dcoupe le rsultat en fonction des retours
%% chariots, et renvoie le reste des envoies  la socket

%% Expliquer le module en deux parties bien distinctes:
%% - La gestion des connexions entrantes et la boucle pour chaque client.
%% - La gestion du protocole entre le client et le serveur avec la rechercher du caractre \n et la ncessit de converver le reste (C'est l qu'il faut expliquer le fonctionnement de TCP/IP.

action(Sock, Donnees) ->
    action(Sock, Donnees, []).

%% Quand les donnes sont traites, on renvoie le reste pour traitement ultrieur, aprs rception ventuelle de donnes complmentaires.
action(Sock, [], Reste) ->
    lists:reverse(Reste);
%% Lorsqu'on reoit le caractere retour chariot, on renvoie l'cho, et on poursuit le traitement.
%% Deux clauses sont ncessaires car le  retour chariot est parfois code '\r\n'
action(Sock, "\r\n" ++ Caracteres, Message) ->
    reponse(Sock, lists:reverse(Message)),
    action(Sock, Caracteres, []);
action(Sock, "\n" ++ Caracteres, Message) ->
    reponse(Sock, lists:reverse(Message)),
    action(Sock, Caracteres, []);

action(Sock, [Caractere|Caracteres], Message) ->
    %% On continue l'accumulation pour rechercher le caractre '/n'.
    action(Sock, Caracteres, [Caractere|Message]).

reponse(Socket, Message) ->
    DateHeure = genere_dateheure(),
    Reponse = io_lib:format("{\"~s\",\"~s\"}\n", [DateHeure, Message]),
    gen_tcp:send(Socket, Reponse).

genere_dateheure() ->
    {{Annee, Mois, Jour}, {Heure, Minute,Seconde}} = calendar:local_time(),
    %% Formattage de la date/heure:
    %% Remarque: Le rsultat de la commande de gnration de la chane
    %% de caractres de date/heure cre une liste que n'est pas
    %% plate. (TODO:ex) Cependant, pour les flux qui sont envoys sur
    %% fichier ou sur socket, il n'est pas ncessaire d'aplatir le
    %% flux. Le fonction de manipulation de flux rseau ou fichier
    %% s'accommode bien de listes ayant une profondeur >1
    io_lib:format("~w~2.2.0w~2.2.0w-~2.2.0w~2.2.0w~2.2.0w",
                  [Annee,
                   Mois,
                   Jour,
                   Heure,
                   Minute,
                   Seconde]).

%% Eventuellement, montrer que l'on peut couper la connexion si l'on essaie d'envoyer un message trop long, en testant la longueur de message
%% Posons par exemple que l'on ne peut traiter de message plus long que 40 caractres.
%% Dans ce cas, on envoie un message au client: "Erreur: Message trop long" et on coupe la connexion.

%% Fin du module echo_server
