%%%-------------------------------------------------------------------
%%% File    : ldap_accept_fsm.erl
%%% Author  : Mickael Remond <mickael.remond@erlang-fr.org>
%%% Description : Dispatch LDAP clients incoming connection
%%%
%%% Created : 27 Feb 2003 by Mickael Remond
%%%                          <mickael.remond@erlang-fr.org>
%%%-------------------------------------------------------------------
-module(ldap_accept_fsm).
-author('mickael.remond@erlang-fr.org').
-vsn(' $Id:$ ').

-behaviour(gen_fsm).

%%--------------------------------------------------------------------
%% Compile time parameters
%%--------------------------------------------------------------------
-define(SERVER, ?MODULE).
%% The client_handling_fun accept the connection socket as argument
%% and return {ok, PID} as result.
-define(CLIENT_HANDLING_FUN, {ldap_connections_sup, start_connection}).

%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/0]).

%% gen_fsm callbacks
-export([init/1, accepting/2, handle_event/3,
         handle_sync_event/4, handle_info/3, terminate/3, code_change/4]).

%% Implementation specific
-export([connect/1, accept/2]).

-record(state, {lsocket, acceptor, number_of_connection}).

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link/0
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link() ->
    gen_fsm:start_link({local, ?SERVER}, ?MODULE, [], []).

%% On new connection
connect(Socket) ->
    gen_fsm:send_event(?SERVER, {connect, Socket}).

%%====================================================================
%% Server functions
%%====================================================================
%%--------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, StateName, StateData}          |
%%          {ok, StateName, StateData, Timeout} |
%%          ignore                              |
%%          {stop, StopReason}                   
%%--------------------------------------------------------------------
init([]) ->
    %% TODO: If the socket is already binded or no right => need gracefull
    %% error management.
    Port = ldap_proxy_config:proxy_port(),
    {ok, LSocket} = gen_tcp:listen(Port, [binary, 
                                           {packet, asn1},
                                           {active, once},
                                           {reuseaddr, true}]),
error_logger:info_msg("LDAP proxy listening on port ~w.~n", [Port]),

    %% And accepting new connections
    AcceptorPid = spawn_link(?MODULE, accept,
                                   [LSocket, ?CLIENT_HANDLING_FUN]),
    {ok, accepting, #state{lsocket = LSocket, acceptor = AcceptorPid}}.

%%--------------------------------------------------------------------
%% Func: StateName/2
%% Returns: {next_state, NextStateName, NextStateData}          |
%%          {next_state, NextStateName, NextStateData, Timeout} |
%%          {stop, Reason, NewStateData}                         
%%--------------------------------------------------------------------
accepting({connect, Pid}, StateData) ->
    %% Acceptor process: We could check here that we did not reach the
    %% max number of connection

    %% Accept new connection
    AcceptorPid = spawn_link(?MODULE, accept,
                             [StateData#state.lsocket, ?CLIENT_HANDLING_FUN]),
    
    {next_state, accepting, StateData#state{acceptor = AcceptorPid}}.

%%--------------------------------------------------------------------
%% Func: handle_event/3
%% Returns: {next_state, NextStateName, NextStateData}          |
%%          {next_state, NextStateName, NextStateData, Timeout} |
%%          {stop, Reason, NewStateData}                         
%%--------------------------------------------------------------------
handle_event(Event, StateName, StateData) ->
    {next_state, StateName, StateData}.

%%--------------------------------------------------------------------
%% Func: handle_sync_event/4
%% Returns: {next_state, NextStateName, NextStateData}            |
%%          {next_state, NextStateName, NextStateData, Timeout}   |
%%          {reply, Reply, NextStateName, NextStateData}          |
%%          {reply, Reply, NextStateName, NextStateData, Timeout} |
%%          {stop, Reason, NewStateData}                          |
%%          {stop, Reason, Reply, NewStateData}                    
%%--------------------------------------------------------------------
handle_sync_event(Event, From, StateName, StateData) ->
    Reply = ok,
    {reply, Reply, StateName, StateData}.

%%--------------------------------------------------------------------
%% Func: handle_info/3
%% Returns: {next_state, NextStateName, NextStateData}          |
%%          {next_state, NextStateName, NextStateData, Timeout} |
%%          {stop, Reason, NewStateData}                         
%%--------------------------------------------------------------------
handle_info(Info, StateName, StateData) ->
    {next_state, StateName, StateData}.

%%--------------------------------------------------------------------
%% Func: terminate/3
%% Purpose: Shutdown the fsm
%% Returns: any
%%--------------------------------------------------------------------
terminate(Reason, StateName, StatData) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change/4
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState, NewStateData}
%%--------------------------------------------------------------------
code_change(OldVsn, StateName, StateData, Extra) ->
    {ok, StateName, StateData}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
%% Accept new connection and spawn client handling fun
accept(LSocket, ClientHandlingFun) ->
    {ok, Socket} = gen_tcp:accept(LSocket),
    error_logger:info_msg("New incoming connection (Socket ~p)~n", [Socket]),

    %% Spawn a process to handle the incoming connection
    Pid = handle_connection(ClientHandlingFun, Socket),

    %% Place the acceptor in a position where it can accept new connection.
    ?MODULE:connect(Pid).

%% Manage synchronization of the incoming connection
%% and the server connection
handle_connection(ClientHandlingFun, Socket) ->
    {Module, Function} = ClientHandlingFun,
    
    case apply(Module, Function, [Socket]) of
        %% Connection established with the server
        {ok, ClientHandlingProcess} ->
            ok = gen_tcp:controlling_process(Socket, ClientHandlingProcess),
            ClientHandlingProcess;
        %% Cannot connect to the server
        {error, Reason} ->
            undefined
    end.
