%%%-------------------------------------------------------------------
%%% File    : ldap_connection_srv.erl
%%% Author  : Mickael Remond <mremond@mremond.cgey.ifp.fr>
%%% Description : 
%%%
%%% Created :  5 Mar 2003 by Mickael Remond <mremond@mremond.cgey.ifp.fr>
%%%-------------------------------------------------------------------
-module(ldap_connection_srv).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/1]).

%% gen_server callbacks
-export([init/1,
         handle_call/3,
         handle_cast/2,
         handle_info/2,
         terminate/2,
         code_change/3]).

-record(state, {client_socket, server_socket}).
-define(SERVER, ?MODULE).

%% LDAP/ASN1 related includes
-include("ELDAPv3.hrl").

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link/1
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link(Client_Socket) ->
    %% Get LDAP server and port to proxy:
    ServerHost = ldap_proxy_config:server_host(),
    ServerPort = ldap_proxy_config:server_port(),
    io:format("Lancement connection sur: ~p:~p~n", [ServerHost, ServerPort]),

    %% Start the connection with the server
    Result = gen_tcp:connect(ServerHost,
                             ServerPort,
                             [binary,
                              {packet, asn1},
                              {active, once}]),

    io:format("Connection au serveur:~p~n", [Result]),

    %% Handle possible errors and do post connection tasks
    post_connection(Result, Client_Socket).


%%====================================================================
%% Server functions
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init/1
%% Description: Initiates the server
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%--------------------------------------------------------------------
init([State]) ->
    {ok, State}.

%%--------------------------------------------------------------------
%% Function: handle_call/3
%% Description: Handling call messages
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_call(Request, From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
%% Function: handle_cast/2
%% Description: Handling cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
handle_cast(Msg, State) ->
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info/2
%% Description: Handling all non call/cast messages
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%--------------------------------------------------------------------
%%% Get data from the server or from the client
handle_info({tcp, Socket, Data}, State) ->
    ServerSocket = State#state.server_socket,
    ClientSocket = State#state.client_socket,

    %% Get module in charge of storing LDAP data that transit through the proxy
    Module    = ldap_proxy_config:proxy_handler(),
    ExtraArgs = ldap_proxy_config:extra_args(),

    Message = asn1rt:decode('ELDAPv3', 'LDAPMessage', Data),
    Side = case Socket of
        ServerSocket -> gen_tcp:send(ClientSocket, Data),
                        inet:setopts(ServerSocket, [{active, once}]), %% Ready to receive a new message on the server socket
                        "server";
        ClientSocket -> gen_tcp:send(ServerSocket, Data),
                        inet:setopts(ClientSocket, [{active, once}]), %% Ready to receive a new message on the client socket
                        "client"

    end,
    Module:log(ServerSocket, ClientSocket, Side, Message, ExtraArgs),
    {noreply, State};

%%% When the client or the server disconnect:
%%% Close the other socket and terminate the process normally
handle_info({tcp_closed, Socket}, State) ->
    ServerSocket = State#state.server_socket,
    ClientSocket = State#state.client_socket,
    case Socket of
        ServerSocket -> gen_tcp:close(ClientSocket);
        ClientSocket -> gen_tcp:close(ServerSocket)
    end,
    error_logger:info_msg("Connection closed (~p)~n", [ClientSocket]),
    {stop, normal, State}.

%%--------------------------------------------------------------------
%% Function: terminate/2
%% Description: Shutdown the server
%% Returns: any (ignored by gen_server)
%%--------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%--------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%--------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
%%% Connection to the server is ok: 
post_connection({ok, Server_Socket}, Client_Socket) ->
    %% Generate server state
    State = #state{client_socket = Client_Socket,
                   server_socket = Server_Socket},

    %% Start Client/Server proxy process and transfer control of the socket.
    {ok, ConnectionManagerPID} = gen_server:start_link(?MODULE, [State], []),
    ok = gen_tcp:controlling_process(Server_Socket, ConnectionManagerPID),

    {ok, ConnectionManagerPID};
%% Connection to the server fails:
%% Handles the error
post_connection({error, Reason}, Client_Socket) ->
    error_logger:info_msg("Cannot connect to server ~s:~w. Reason = ~p~n",
                          [ldap_proxy_config:server_host(),
                           ldap_proxy_config:server_port(),
                           Reason]),
    {error, Reason}.

