%%%-------------------------------------------------------------------
%%% File    : rei_server_accept_fsm
%%% Author  : Mickael Remond <mickael.remond@erlang-fr.org>
%%% Description : Acceptor, dispatcher for incoming connexion
%%% Created : 22 Nov 2002 by Mickael Remond <mickael.remond@erlang-fr.org>
%%%-------------------------------------------------------------------
-module(rei_server_accept_fsm).
-author("mickael.remond@erlang-fr.org").

-behaviour(gen_fsm).

%% External exports
%%%  Connect and disconnect are taking the socket as parameter
-export([start_link/0, connect/1, disconnect/1]).

%% gen_fsm callbacks
-export([init/1, accepting/2, handle_event/3,
	 handle_sync_event/4, handle_info/3, terminate/3, code_change/4]).

%% Implementation specific
-export([accept/1]).

-record(state, {lsocket, acceptor, number_of_connexions = 0}).

-define(PORT, 1666).
-define(SERVER, acceptor).

%%====================================================================
%% Server functions
%%====================================================================
start_link() ->
    gen_fsm:start_link({local, ?SERVER}, ?MODULE, [], []).

%% On new connexion
connect(Socket) ->
    gen_fsm:send_event(?SERVER, {connect, Socket}).

%% On disconnect
disconnect(Socket) ->
    gen_fsm:send_event(?SERVER, {disconnect, Socket}).    

%%====================================================================
%% Server functions
%%====================================================================
init([]) ->
    Opts =
        [{active, once},
         {nodelay, true},
         {packet,raw},
         {reuseaddr,true}],

    % We crash if we can't open the socket to listen
    {ok, ListenSocket} = gen_tcp:listen(?PORT, Opts),

    error_logger:info_msg("Rei Server listening on port ~w.~n", [?PORT]),

    %% Accepting new connexions
    AcceptorPid = spawn_link(?MODULE, accept,
				   [ListenSocket]),
    {ok, accepting, #state{lsocket = ListenSocket, acceptor = AcceptorPid}}.

accepting({connect, Pid}, StateData) ->
    %% Acceptor process: We could check here that we did not reach the
    %% max number of connexion
    NumConnexions = StateData#state.number_of_connexions,

    %% Accept new connexion
    AcceptorPid = spawn_link(?MODULE, accept, [StateData#state.lsocket]),
    
    {next_state, accepting, StateData#state{acceptor = AcceptorPid,
					    number_of_connexions=NumConnexions + 1}};

accepting({disconnect, Pid}, StateData) ->
    NumConnexions = StateData#state.number_of_connexions,

    {next_state, accepting, StateData#state{number_of_connexions=NumConnexions - 1}}.



%% This is unused part of the gen_fsm behaviour interface
handle_event(Event, StateName, StateData) ->
    {next_state, StateName, StateData}.
handle_sync_event(Event, From, StateName, StateData) ->
    Reply = ok,
    {reply, Reply, StateName, StateData}.
handle_info(Info, StateName, StateData) ->
    {next_state, StateName, StateData}.
terminate(Reason, StateName, StatData) ->
    ok.
code_change(OldVsn, StateName, StateData, Extra) ->
    {ok, StateName, StateData}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
%% Accept new connexion and spawn client handling fun
accept(LSocket) ->
    {ok, Socket} = gen_tcp:accept(LSocket),
    error_logger:info_msg("New incoming connection (Socket ~p)~n", [Socket]),

    %% Spawn a process to handle the incoming connexion
    Pid = manage_connection(Socket),

    %% Register the new process in the notify server 
    rei_server_notify_srv:add(Pid),

    %% Place the acceptor in a position where it can accept new connection.
    ?MODULE:connect(Pid).

%% Manage synchronization of the incoming connection
%% and the server connection
manage_connection(Socket) ->
    case supervisor:start_child(rei_server_connect_sup, [Socket]) of
	%% Connection established with the server
	{ok, ClientHandlingProcess} ->
	    ok = gen_tcp:controlling_process(Socket, ClientHandlingProcess),
	    ClientHandlingProcess;
	%% Cannot connect to the server
	{error, Reason} ->
	    undefined
    end.

