%%%-------------------------------------------------------------------
%%% File    : rei_server_dispatch_srv.erl
%%% Author  : Mickael Remond <mickael.remond@erlang-fr.org>
%%% Description : Handle each client connection
%%%
%%% Created : 23 Nov 2002 by Mickael Remond
%%%                          <mickael.remond@erlang-fr.org>
%%%-------------------------------------------------------------------
-module(rei_server_dispatch_srv).

-behaviour(gen_server).
%%--------------------------------------------------------------------
%% Include files
%%--------------------------------------------------------------------

%%--------------------------------------------------------------------
%% External exports
-export([start_link/1, send/2, close/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

%% Socket is the client socket
-record(state, {socket, data_buffer=[]}).

%%====================================================================
%% External functions
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link/0
%% Description: Starts the server

%%--------------------------------------------------------------------
start_link(Socket) ->
    gen_server:start_link(?MODULE, [Socket], []).

send(Pid, Message) ->
    gen_server:cast(Pid, {send, Message}).

close(Pid) ->
    gen_server:cast(Pid, {close}).

%%====================================================================
%% Server functions
%%====================================================================

init([Socket]) ->
    io:format("New connection~n", []),
    {ok, #state{socket=Socket}}.

%% Close the socket and stop the server
handle_cast({close}, State) ->
    Socket = State#state.socket,
    gen_tcp:close(Socket),
    {stop, normal, State};
%% Send data to the client
handle_cast({send,Message}, State) ->
    Socket = State#state.socket,
    gen_tcp:send(Socket, Message),
    {noreply, State}.

handle_info({tcp,Socket,Data}, State) ->
    inet:setopts(Socket, [{active, once}]), %% Ready to receive a new message on the server socket
    {Lines, RemainingData} = process_data(State#state.data_buffer ++ Data),
    lists:foreach(fun(Line) -> process_line(Line, Socket) end, Lines),
    {noreply,State#state{data_buffer = RemainingData}};
handle_info({tcp_closed,Socket}, State) ->
    error_logger:info_msg("Client ~p disconnected~n", [Socket]),
    {stop, normal, State};
handle_info(Info, State) ->
    io:format("Info: ~p~n", [Info]),
    {noreply,State}.

%% Unused gen_server function API
handle_call(Request, From, State) ->
    Reply = ok,
    {reply, Reply, State}.
terminate(Reason, State) ->
    ok.
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
%----------------------------------------------------------------------
%% process_data/5
%% Parse the incoming data into lines,
%%----------------------------------------------------------------------
process_data(Data) ->
    process_data(Data, []).
process_data(Data, Lines) -> 
    Pos = string:str(Data, "\n"),
    case Pos of 
        0 -> {Lines, Data};
        _ ->
            Line = get_line(Data, Pos),
            Rest = string:substr( Data, Pos+1, string:len(Data) - (Pos-1) ), 
            process_data(Rest, Lines ++ [Line])
    end.

process_line([], Socket) ->
    ok;
process_line(Line, Socket) ->
    [Command | Params] = string:tokens(Line, [$ ]),
    %case apply(list_to_atom(httpd_util:to_lower(Command)),
    Module = list_to_atom(httpd_util:to_lower(Command)),
    case catch Module:process(self(), Params) of
        {'EXIT', Error} ->
            error_logger:info_msg("Protocol error: ~p~n     -> ~p~n", [Command, Error]),
            gen_tcp:send(Socket, "ERROR\r\n");
        {continue, Response} ->
            gen_tcp:send(Socket, Response);
        {close, Response} ->
            gen_tcp:send(Socket, Response),         
            close(self())
    end.

%% Return a complete command line, given the position of the \n
%% character
get_line(Data, Pos) ->
    Line  = string:substr( Data, 1, Pos-1 ),
    %% Remove trailing \r characters if any
    Line2 = string:strip(Line, right, $\r),
    Line2.
