-module(wpc_noise).
%% Extension pour le modeleur Wings 3D
%%  Mickal Rmond <mickael.remond@erlang-fr.org>
%%
%% Ce module propose deux types de fonctionnalits:
%%  - propose des fonctions d'aide permettant d'afficher les
%%    informations d'tat de la scne.
%%    Le code permet d'illustrer l'accs aux valeurs d'tat depuis un
%%    module d'extension. Ces fragments de code sont directement
%%    utilisables dans vos propres modules d'extension.
%%  - propose un exemple de modification simple de la scne en cours.
%%    Le module ajoute du "bruit"  la scne courante en dplaant
%%    alatoirement tous les points de la scne

%% Compilation
%%   $ erlc -I ../../e3d/ -I ../../src/ wpc_debug2.erl

%% API d'une extension Wings
-export([init/0,menu/2,command/2]).

%% Il faut toujours inclure ces fichiers dans les extensions  Wings
%% 3D
-include_lib("esdl/include/gl.hrl").
-include("e3d.hrl").
-include("e3d_image.hrl").
-include("wings.hrl").

%% On profite d'init pour initialiser le gnrateur de nombres
%% alatoires
init() ->
    {A,B,C} = erlang:now(),
    random:seed(A,B,C),
    true.


%%	      -=-=-=-=- Paramtrage des menus -=-=-=-=-
%% Ajout d'options  la fin du menu d'aide
menu({help}, Elements) ->
    Elements ++ [separator,
		 {"Informations gnrales", general_info},
		 {"Liste des objets", objects},
		 {"Liste des slections", select_list}];

%% Ajout d'une option alatoire dans le menu outil
menu({tools}, Elements) ->
    Elements ++ [separator,
		 {"Changements alatoires", random}];

%% Les autres menus sont ignors
menu(_EntreeDeMenu, _Elements) ->
    _Elements.


%%	 -=-=-=- Affichage des informations gnrales -=-=-=-
%% Gnration d'un cran d'informations gnrales  partir de l'tat
%% du monde
command({help, general_info}, State) ->
    %% Rcupre le nom du fichier en cours d'dition (#st.file)
    NomFichier = case State#st.file of
		     undefined -> "Noname";
		     Fichier   -> Fichier
		 end,

    %% Rcupre le mode de slection actif dans Wings (#st.selmode)
    SelMode = atom_to_list(State#st.selmode),

    %% Rcupre l'tat de l'option smart highlight (#st.sh)
    SmartHighlight = atom_to_list(State#st.sh),

    %% Met en forme l'cran
    wpa:dialog("Informations gnrales",
	       [label("Fichier: " ++ NomFichier),
		label("Mode de slection: " ++ SelMode),
		label("Smart Highlight: " ++ SmartHighlight)],
	       %% Evnment  gnrer  la sortie de l'cran:
	       fun(Resultat) -> {help,{general_info, Resultat}} end);
%% La sortie de l'cran d'information ne ncessite pas de validation.
%% C'est l'endroit o il faut habituellement insrer contrles et
%% traitements
command({help, {general_info, Resultat}}, State) ->
    keep;

%%	   -=-=-=- Affichage de la liste des objets -=-=-=-
command({help, objects}, State) ->
    Formes = gb_trees:keys(State#st.shapes),


    %% Prpare une liste de labels pour l'affichage des objets de la scne
    Recap = [label(io_lib:format("~w formes:", [length(Formes)]))],
    Objets = lists:map(fun(FormeId) ->
			       WingedEdge = gb_trees:get(FormeId, State#st.shapes),
			       Nom = WingedEdge#we.name,
			       label(io_lib:format(" - Forme ~w: ~s", [FormeId, Nom]))
		  end,
		  Formes),

    %% Met en forme l'cran
    wpa:dialog("Liste des objets",
	       Recap ++ Objets,
	       fun(Resultat) -> {help,{objects,Resultat}}
	       end);
command({help, {objects, Resultat}}, State) ->
    keep;

%%  -=-=-=- Affichage des informations sur les slections -=-=-=-
command({help, select_list}, State) ->
    %% Rcupre l'tat de la slection courante (#st.sel)
    Selection = State#st.sel,

    %% Prpare une liste de labels pour l'affichage des slections de la scne
    RecapSelCourante = [label(io_lib:format("~w objet(s) dans selection courante:",
					    [length(Selection)]))],

    SelCourante = lists:map(fun({ObjetSelId, ElementSel}) ->
				    [label(io_lib:format(" * Objet ~w (~w lments):",
							 [ObjetSelId, gb_sets:size(ElementSel)]))] ++
					selection(ElementSel)
			    end,
			    Selection),

    %% Rcupre le nom et le mode des slections mmorises existantes
    %% (#st.ssels)
    GroupeIds = gb_trees:keys(State#st.ssels),
    RecapGroupes = [label(io_lib:format("~w groupes mmorises:", [length(GroupeIds)]))],
    Groupes = lists:map(fun({ModeSel, Nom}) ->
				label(io_lib:format(" - ~s (mode: ~p)", [Nom, ModeSel]))
			end,
			GroupeIds),

    %% Met en forme l'cran
    wpa:dialog("Slections",
	       lists:flatten(RecapSelCourante ++ SelCourante ++ [label("--")] ++
	       RecapGroupes ++ Groupes),
	       fun(Resultat) -> {help,{objects,Resultat}}
	       end);
command({help, {select_list, Resultat}}, State) ->
    keep;

%%% La version peut tre toffe sur tous les lments de l'tat
% Affichage des textures dfinies (#st.mat)
% A rintgrer:
%     TextureIds = gb_trees:keys(State#st.mat),
%     io:format("~w Textures:~n", [length(TextureIds)]),
%     lists:foreach(fun(Result) ->
% 			  io:format(" - ~p~n", [Result])
% 		  end,
% 		  TextureIds),

%%	   -=-=- Transformation de la scne courante -=-=-
command({tools, random}, State) ->
    %% Rcupre les identifiants des formes
    Shapes = State#st.shapes,
    FormeIds = gb_trees:keys(Shapes),

    %% Pour chacune des formes rcupre les points et modifie alatoirement les coordonnes
    Result = lists:map(fun(FormeId) ->
			       WingedEdge = gb_trees:get(FormeId, Shapes),
			       Coordonnees = gb_trees:to_list(WingedEdge#we.vp),
			       {FormeId, WingedEdge#we{vp = gb_trees:from_orddict(lists:keysort(1,randomize_we(Coordonnees,[])))}}
		  end,
		  FormeIds),
    NewShapes = Shapes,
    State#st{shapes = gb_trees:from_orddict(lists:keysort(1,Result))};

%% Les autres vnements doivent tre grs par Wings eux-mme ou par
%% d'autres modules.
command(_Event, _State) ->
    next.


%% -=-=- Fonctions de mise en forme des crans d'aides -=-=-
%% label/1 permet de gnrer le tuple attendu par wpa:dialog/3
%%% Attention: Il faut utiliser lists:flatten/1, car la fonction de
%%% cration de l'interface de dialogue ne fonctionne bien qu'avec des
%%% listes plates
%%% io_lib:format/2 ne gnre pas une liste plate
label(Message) ->
    {label, lists:flatten(Message)}.

%% Renvoie une chaine dcrivant la slection dans un objet donne
%% Illustration du parcours de gb_sets
selection(GbSet) ->
    Iter = gb_sets:iterator(GbSet),
    selection(Iter, []).
selection(Iter, Liste) ->
    case gb_sets:next(Iter) of
	{Value, Iter2} -> %% Element
	    Element = label(io_lib:format("   - Element ~p~n", [Value])),
	    selection(Iter2, Liste ++ [Element]);
	none -> %% Plus d'lements
	    Liste
    end.

%% Traitement pour chacun des points
randomize_we([], Acc) ->
    lists:reverse(Acc);
randomize_we([{Id, Coordonnees}|Vertex], Acc) ->
    randomize_we(Vertex, [{Id, noise(Coordonnees)}|Acc]).
%% Changement alatoire des coordonnes d'un point
noise(Coordonnees) ->
    Result = lists:map(
	       fun(Valeur) ->
		       %% Sens de la modification sur l'axe
		       Sens = random:uniform(3) - 2,
		       %% Ampleur (Jusqu' 15 % de la coordonne)
		       %% Algorithme brutal pouvant tre amlior
		       %% Exercice pour le lecteur: Rendre l'ampleur
		       %% de la modification paramtrable
		       Ampleur = random:uniform(15),
		       Valeur + (Sens * (Ampleur * Valeur/100))
	       end,
	       tuple_to_list(Coordonnees)),
    list_to_tuple(Result).
