-module(hello_srv).

%% Dclaration d'utilisation du comportement
-behaviour(gen_server).

%% fonctions requises par le comportement gen_server
%% (callbacks). handle_cast/2
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

%% fonctions de contrle du serveur
-export([start/0, start_link/0, stop/0]).

%% fonctions de stockage et rcupration des valeurs
-export([store/2, retrieve/1]).

%% La boucle du serveur lui-mme n'existe plus car son code est
%% prsent dans le gen_server.

%% Macro dfinissant le nom du serveur 
%% Dfinir ce nom  l'aide d'une macro permet d'en changer trs
%% facilement.
-define(server, hellosrv).

%% Nous pouvons optionnellement dfinir un record pour grer l'tat du
%% serveur.

%% ----------------------------------------
%% Interface du serveur (API)
%% Implmentation des fonctions clientes.

%% start: Dmarrage du serveur.
%% Le premier paramtre permet de nommer le serveur sur le noeud local
%% et d'utiliser ce nom pour toutes les communications avec le
%% processus du serveur
start() ->
    gen_server:start({local, ?server}, ?MODULE, [], []).

%% La fonction start_link n'est pas obligatoire. Elle permet
%% simplement par la suite d'intgrer notre serveur dans une
%% application OTP complte.
%% Elle le le pendant de la fonction start.
start_link() ->
    gen_server:start_link({local, ?server}, ?MODULE, [], []).

%% Stop: arrte le serveur 
%% On utilise le nom du serveur
stop() ->
    gen_server:call(?server, stop, 10000).

%% Les fonctionnalits du serveur = son API.

%% enregistre
%%  Pid = Identifiant de processus
%%  Cle = chane de caractres
%%  Valeur = chane de caractres
store(Cle, Valeur) ->
    %% Le passage de la rfrence de processus du client n'est plus ncessaire.
    %% Tout cela est gr par le comportement gen_server.
    gen_server:call(?server, {store, Cle, Valeur}, 10000).
    %% La rcupration de la rponse est galement gre par le gen_server.

%% recupere
%%  Pid = Identifiant de processus
%%  Cle = Chane de caractre.
retrieve(Cle) ->
    gen_server:call(?server, {retrieve, Cle}, 10000).

%% Les fonctions ncessaires au comportement gen_server
%% Il s'agit de l'implmentation du serveur lui-mme

init(Args) ->
    %% La gestion du signal 'EXIT' est imprative si l'on souhaite
    %% conserver l'tat du systme entre les redmarrage effectu par
    %% le superviseur
    %%process_flag(trap_exit, true),

    %% Le dmarrage s'effectue correctement;
    %% L'tat initial du serveur est vide.
    {ok, Args}.

%% Implmentation des traitements correspondant aux messages reus par
%% le serveur.
%%  -*- store -*-
handle_call({store, Cle, Valeur}, _From, State) ->
    NewState = case lists:keymember(Cle, 1, State) of
		   %% Cre l'entre dans la liste
		   false -> State ++ [{Cle, Valeur}];
		   %% Remplace l'entre existante
		   true -> lists:keyreplace(Cle, 1, State, {Cle, Valeur})
	       end,
    %% La rponse doit rpondre  un formalisme particulier.
    %% Ici, le traitement s'est bien droul et l'tat du serveur est
    %% mis  jour.
    {reply, ok, NewState};
%%  -*- retrieve -*-
handle_call({retrieve, Cle}, _From, State) ->
    case lists:keysearch(Cle, 1, State) of
	%% Nous introduisons un bug ici en commentant cette ligne
	%% Lorsque nous cherchons une valeur qui n'existe pas, nous obtenons
	%% une erreur de clause.
	%false -> {reply, {error, not_found}, State};
	%% La cl a t trouve:
	{value, {Cle, Valeur}} -> {reply, {ok, Valeur}, State}
    end;
%%  -*- stop -*-
handle_call(stop, _From, State) ->
    {stop, normal, ok, State}.

%% Aucun traitement asynchrone n'est prvu: Le traitement ne fait rien.
handle_cast(_Request, State) ->
    {noreply, State}.

%% Aucun traitement n'est prvu en cas de demande d'information
%% (Signal 'EXIT' ou messages directs au processus)
handle_info(_Request, State) ->
    io:format("Info: Reu: ~p~n", [_Request]),
    {noreply, State}.    

%% Cette fonction sert  dfinir les traitements  effectuer lors de
%% l'arrt du serveur.
%% On peut par exemple en profiter pour sauvegarder l'tat du serveur
%% pour dboguage ou bien prvenir l'administrateur du systme par
%% SMS.
terminate(Reason, State) ->
    ok.

%% Aucun traitement de mise  jour des donnes (Etat du serveur)
%% n'est pour le moment prvue.
%% L'tat du serveur est repris tel quel. Nous pouvons considrer pour
%% le moment qu'une mise  jour de code ne modifie pas la manire dont
%% le serveur gre ses donnes en interne.
code_change(OldVsn, State, Extra) ->
    State.
