% Interface module (UI) for Game Of Life (GOL) in Erlang
%
% Author:		Stphane Perrot (s.perrot@voila.fr)
%

-module(gs_ui).
-export([start/4, init/2]).

-define (BGCOL,forestgreen).
-define (CELLSIZE,60).
-define (DEBUG,false).
-define (TITLE,"Game of Life").

% main entry point for spawn
start (NbrOfCells, 0, Matrix = {N,P}, DispatcherPid) ->
    purge_old_messages(),
    Frame = init(N,P),
    setup_matrix (Frame, N, P),
    loop (DispatcherPid, NbrOfCells, NbrOfCells, [], Frame).

%% Graphics setup
init (N,P) ->
    S = gs:start(),

    %% Shell coordinates
    FrameHeight =  N*?CELLSIZE,
    TotalHeight = 50 + FrameHeight,
   
    Width  = P * ?CELLSIZE,
    Shell = gs:create (window, S, [{title,?TITLE}, 
				   {width, Width},
				   {height, TotalHeight}]),
    setup_buttons (Shell,0,0,Width,50),
    Frame = gs:create (frame, Shell,
		       [{x,0}, {y,50},
			{width,Width},
			{height,FrameHeight},
			{bg,white}]),
    gs:config (Shell, {map, true}),
    Frame.

loop (DispatcherPid, NbrOfCells, 0, StatesAlist, Frame) ->
    display_matrix (Frame, StatesAlist),
    % Go on
    loop (DispatcherPid, NbrOfCells, NbrOfCells, [], Frame);

loop (DispatcherPid,NbrOfCells, Remaining, StatesAlist, Frame) ->
    receive
	{gs,ButtId,click,ButtId1,["Quit"]} ->
	    %%io:format("ui: sending `terminate' message to dispatcher...~n", []),
	    DispatcherPid ! {terminate},
	    ok;

	{gs,ButtId,click,ButtId1,["NewGen"]} ->
	    %%io:format("ui: `nextgen' message sent to dispatcher...~n", []),
	    DispatcherPid ! {nextgen},
	    loop (DispatcherPid, NbrOfCells, Remaining,
		  StatesAlist, Frame);

	{cell, Coord, Generation, CellState} ->
	    %io:format("msg cell received...~n", []),
	    loop (DispatcherPid, NbrOfCells, Remaining - 1,
		 [{Coord, CellState} | StatesAlist], Frame)

    after 30000 ->
	    io:format("ui: no activity, stopping listening...~n"),
	    ok
    end.

% =====================================================================	
setup_buttons (Shell,X,Y,W,H) ->
    ButBox = gs:create (frame, Shell,[{x,X}, {y,Y},{bg,white},
				      {width,W}, {height,H}]),
    ButtonWidth = W / 2 - 10,
    C = gs:create (canvas,ButBox,[{x,X}, {y,Y}, {width, W}, {height, H},
				  {bg,white}]),
    gs:create (line, C, [{coords, [{0,H-1}, {W,H-1}]}]),

    mk_button("Quit",	ButBox, 10,	     10, ButtonWidth, 25),
    mk_button("NewGen",	ButBox, ButtonWidth, 10, ButtonWidth, 25),
    done.

mk_button (Label, Parent, X, Y, W, H) ->
    gs:create(button,Parent,[{width,W}, {height, H}, {x,X}, {y,Y}, 
			     {label, {text, Label}}, {bg,white},
			     {activefg,white}, {activebg,black}]).

% ---------------------------------------------------------------
bitmap (grey)  -> "./bitmaps/square.bm";
bitmap (black) -> "./bitmaps/marker.bm";
bitmap (white) -> "./bitmaps/marker.bm".

coord_to_name(I,J) ->
    integer_to_list(I) ++ "," ++ integer_to_list(J).

xy_position (I,J) ->
    X = ?CELLSIZE * (J-1),
    Y = ?CELLSIZE * (I-1),
    {X,Y}.

mk_board_butt (Top,Name, X,Y,Colour) ->
    Button = gs:create (button,list_to_atom(Name), Top,
		  [{x,X}, {y,Y},
		   {width,?CELLSIZE},
		   {height,?CELLSIZE},
		   {padx,5}, {pady,5},
		   {relief,flat},
		   {bg,		?BGCOL},
		   {activebg,	?BGCOL}]),
    butbit (Button,Colour),
    Button.

butbit (Button,Colour) ->
    gs:config(Button,[{label,{image,bitmap(Colour)}},
		      {fg, Colour},
		      {activefg,Colour},
		      {label, {image, bitmap(Colour)}}]),
    Button.

setup_matrix (Frame, N, P) ->
    List = [{I,J} || I <- lists:seq(1,N),
		     J <- lists:seq(1,P)],
    setup_matrix (Frame, List).

setup_matrix (Frame, []) ->
    true;
setup_matrix (Frame, [Cell | Rest]) ->
    {I,J} = Cell,
    {X,Y} = xy_position (I,J),
    Name = coord_to_name (I,J),
    Button = mk_board_butt (Frame,Name,X,Y,grey),
    put (Name,Button),
    setup_matrix (Frame,Rest).

% ------------------------------------------------------
display_matrix (Frame, []) ->
    true;
display_matrix (Frame, [Cell | Rest]) ->
    {{I,J}, State} = Cell,
    {X,Y} = xy_position(I,J),
    case State of 
	alive -> Colour = white;
	dead  -> Colour = black
    end,
    Name = coord_to_name (I,J),
    Button=get(Name),
    butbit(Button, Colour),
    display_matrix (Frame,Rest).

% -----------------------------------------------------------------------
purge_old_messages () ->
    %%io:format("purging old messages..."),
    do_purge (500).
    %%io:format("done~n").

do_purge (Delay) ->
    receive
	{cell, _, _, _}	-> do_purge (Delay);
	{gs,_,_,_,_}	-> do_purge (Delay);
	{nextgen}	-> do_purge (Delay)
    after Delay		-> ok
    end.
% -----------------------------------------------------------------------
